/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>                          /* fcntl() */
#include "initng.h"
#include "initng_global.h"
#include "initng_service_cache.h"
#include "initng_load_module.h"
#include "initng_toolbox.h"
#include "initng_signal.h"

#include "initng_plugin_callers.h"
#include "initng_plugin.h"

#include "initng_fd.h"

void initng_close_all_fd(void)
{
    s_call *current, *safe = NULL;

    S_;

    while_call_db_safe(current, safe)
    {
        if (current->ptype != FDWATCHERS)
            continue;

        close(current->c.fdh->fds);
        list_del(&current->list);

        if (current->from_file)
            free(current->from_file);
        free(current);
    }
}

/* Calls plugins when data is ready in pipe */
void initng_plugin_readpipe(active_h * service, int datalen,
                            process_h * process, int flush_buffer)
{
    s_call *current = NULL;

    S_;

    while_call_db(current)
    {
        if (current->ptype != PIPEWATCHERS)
            continue;

        D_("Calling pipewatcher plugin.\n");
        (*current->c.pipewatcher) (service, datalen, process, flush_buffer);
    }
}

/*
 * This does the actual read from pipe, all checks have been done here.
 *
 */
int initng_process_readpipe_read(active_h * service, process_h * p,
                                 int flush_buffer)
{
    int len = 0;                /* lenght of data read */

    S_;

    /* check if buffer exits, or have to be created */
    if (!p->buffer)
    {
        /* allocate space for the buffer */
        p->buffer = initng_calloc(MAX_BUFFER + 1, sizeof(char));
        p->buffer_pos = 0;
    }

    /* read data from process, and continue again after a interrupt */
    do
    {
        errno = 0;
        /* can max read MAX_BUFFER (once allocated) - len (all currently read) - buffer_pos */
        if ((MAX_BUFFER - len - p->buffer_pos) < 1)
        {
            F_("Cant read more, buffer is out!\n");
            break;
        }
        len +=
            read(p->out_pipe[0], &p->buffer[p->buffer_pos],
                 (MAX_BUFFER - len - p->buffer_pos));
    }
    while (errno == EINTR);

    /* make sure we got something */
    if (len < 1)
        return (len);

    /* if we got something. */

    /* null the end to make sure */
    p->buffer[p->buffer_pos + len] = '\0';

    /* if max buffer is reached, make sure it flushes it */
    if (p->buffer_pos + len >= MAX_BUFFER)
        flush_buffer = TRUE;

    /* let all plugin take part of data */
    initng_plugin_readpipe(service, len, p, flush_buffer);

    /* increase the buffer possision */
    if (flush_buffer)
        p->buffer_pos = 0;
    else
        p->buffer_pos = p->buffer_pos + len;

    /* return lenght of data read */
    return (len);
}

/* This does all the fancy pancy around the read */
void initng_process_readpipe(active_h * service, process_h * p,
                             int flush_buffer)
{
    int len = 0;

    S_;

    /*
     * flush_buffer is set, and readpipe is called just
     * before process are beeing freed, when the pocess
     * is dead.
     */
    if (flush_buffer)
    {
        int fd_flags;

        fd_flags = fcntl(p->out_pipe[0], F_GETFL, 0);
        fcntl(p->out_pipe[0], F_SETFL, fd_flags | O_NONBLOCK);
    }

    /* get data from process, it return lenght of data got */
    len = initng_process_readpipe_read(service, p, flush_buffer);

    if (!flush_buffer)
    {
        /* We have an error other than EAGAIN, set flush_buffer */
        if ((len == -1) && (errno != EAGAIN))
        {
            flush_buffer = TRUE;
            F_("Error reading from pipe for %s (%s)\n", service->name,
               strerror(errno));
        }

        /* Why flush buffer if we got nothing to read? */
        /* End of file */
        if (len == 0)
            flush_buffer = TRUE;
    }

    /* If i < 1 we either have 0 (end of file) or -1 (error) */
    if (!flush_buffer)
        return;

    /*
     * Do this to tell plugin that buffer is flushing,
     * and data is going to dissapere.
     */
    initng_plugin_readpipe(service, 0, p, TRUE);


    /* free data, close and reset fds */
    D_("Freeing buffer for %s\n", service->name);
    if (p->out_pipe[0] > 0)
        close(p->out_pipe[0]);
    if (p->out_pipe[1] > 0)
        close(p->out_pipe[1]);
    if (p->buffer)
        free(p->buffer);
    p->out_pipe[0] = -1;
    p->out_pipe[1] = -1;
    p->buffer = NULL;
}


/*
 * FILEDESCRIPTORPOLLNG
 *
 * Filedescriptors are scaned in 3 steps.
 * First, scan all plugins, and processes after filedescriptors to
 * watch, and add them with FD_SET(); added++;
 *
 * Second, Run select if there where any added, with an timeout set
 * when function was called.
 *
 * Third, if select returns that there is data to fetch, walk over
 * plugins and filedescriptors again, and call plugins or initng_process_readpipe
 * with pointer to plugin/service.
 *
 * This function will return FALSE, if it timeout, or if there was
 * no fds to monitor.
 */
int initng_plugin_fd_poll(int timeout)
{
    /* Variables */
    fd_set readset;
    struct timeval tv;
    int retval;
    int added = 0;
    active_h *currentA, *qA;
    s_call *currentC, *qC;
    process_h *currentP, *qP;

    /* initziation */
    S_;
    FD_ZERO(&readset);

    /* Set timeval struct, to our timeout */
    tv.tv_sec = timeout;
    tv.tv_usec = 0;


    /*
     * STEP 1:  Scan for fds to add
     */


    /* scan through active plugins that have listening file descriptors, and add them */
    currentC = NULL;
    while_call_db(currentC)
    {
        /* first, some checking */
        if (currentC->ptype != FDWATCHERS)
            continue;
        if (currentC->c.fdh->fds <= 2)
            continue;
        if (!currentC->c.fdh->call_module)
            continue;

        /*D_("adding fd #%i, from an call_db.\n", current->c.fdh->fds); */
        FD_SET(currentC->c.fdh->fds, &readset);
        added++;
    }

    /* Then add all file descriptors from process read pipes */
    currentA = NULL;
    while_active_db(currentA)
    {
        currentP = NULL;
        while_processes(currentP, currentA)
        {
            if (currentP->out_pipe[0] > 2)
            {
                FD_SET(currentP->out_pipe[0], &readset);
                added++;
            }
        }
    }


    /*
     * STEP 2: Do the select-poll, if any fds where added
     */

    /* check if there are any set */
    if (added <= 0)
    {
        D_("No filedescriptors set.\n");
        sleep(timeout);
        return (TRUE);
    }
    D_("%i filedescriptors added.\n", added);


    /* If interrupt then return and do main loop, it doesn't hurt if the filebuffers have a bit more data anyway */
    if (g.interrupt)
        return (TRUE);

    /* make the select */
    retval = select(256, &readset, NULL, NULL, &tv);

    /* if interupt is set, don't continue handeling fussy fds */
    if (g.interrupt)
        return (TRUE);

    /* error */
    if (retval < 0)
    {
        F_("Select returned with %d %s\n", retval, strerror(errno));
        return (FALSE);
    }

    /* none was found */
    if (retval == 0)
    {
        D_("There was no data found on any added fd.\n");
        return (FALSE);
    }

    D_("%d fd's active\n", retval);

    /* If a fsck is running select will always return one filehandler active, we give it 0.1 seconds to get some more data into the buffer. */
    sleep(0.1);


    /*
     * STEP 3:  Find the plugins/processes that handles the fd input, and call them
     */

    /* Now scan throught callers */
    currentC = NULL;
    qC = NULL;
    while_call_db_safe(currentC, qC)
    {
        if (currentC->ptype != FDWATCHERS)
            continue;
        if (currentC->c.fdh->fds <= 2)
            continue;
        if (!currentC->c.fdh->call_module)
            continue;
        if (!FD_ISSET(currentC->c.fdh->fds, &readset))
            continue;

        D_("Calling pluginhandler for fd #%i\n", currentC->c.fdh->fds);
        (*currentC->c.fdh->call_module) (currentC->c.fdh);
        D_("pluginhandler for fd #%i returned!\n", currentC->c.fdh->fds);

        /* Found match, that means we need to look for one less, if we've found all we should then return */
        retval--;
        if (retval == 0)
            return (TRUE);
    }

    /* scan every service */
    currentA = NULL;
    qA = NULL;
    while_active_db_safe(currentA, qA)
    {
        currentP = NULL;
        qP = NULL;
        /* and all the processes */
        while_processes_safe(currentP, qP, currentA)
        {
            /* if matching */
            if (currentP->out_pipe[0] > 2
                && FD_ISSET(currentP->out_pipe[0], &readset))
            {
                D_("Will read from %s->start_process on fd #%i\n",
                   currentA->name, currentP->out_pipe[0]);
                /* Do the actual read from pipe */
                initng_process_readpipe(currentA, currentP, FALSE);
                /* If we've been interrupted, return to main loop (program might have ended) */
                if (g.interrupt)
                    return (TRUE);
                /* Found match, that means we need to look for one less, if we've found all we should then return */
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
        }
    }

    return (TRUE);
}
